package net.pricefx.common.apibuilder.calculationflowaction;

import net.pricefx.common.apibuilder.Invocations;
import net.pricefx.common.dto.ChainedJobInfo;

import java.util.List;

/**
 * This action starts calculation of selected CFS. It's a programmatic version
 * of the same action available in UI. As such it allows user to schedule the
 * action in calculation flow formula.
 */
public class CFSAction<ATTACHEDTO extends Help> extends Action {

	private String cfsLabel;
	private Boolean calculate;
	private int priority;
	private int distributedCalculationThreshold;
	private ATTACHEDTO backLink;
	
	public CFSAction(ATTACHEDTO backLink, Invocations<MethodInvocationRecord> invocations, String cfsLabel) {
		super(invocations);
		this.backLink = backLink;
		this.cfsLabel = cfsLabel;
		this.calculate = false;
	}

	/**
	 * Whether the action is valid and should be triggered or not.
	 * 
	 * @param calculate
	 *            <ul>
	 *            <li>true - calculation of the CFS is started as background
	 *            process</li>
	 *            <li>false - calculation of the CFS is NOT started at all</li>
	 *            </ul>
	 */
	public CFSAction<ATTACHEDTO> setCalculate(Boolean calculate) {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".setCalculate", calculate));
		this.calculate = Boolean.TRUE.equals(calculate) ? true : false;
		return this;
	}
	
	/**
	 * Sets the prority of this job. 
	 * 
	 * @param priority Default is 0 (lowest user-definable priority). Max value is 9.
	 * @return
	 */
	public CFSAction<ATTACHEDTO> setPriority(int priority) {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".setPriority", priority));
		this.priority = priority;
		return this;		
	}		

	/**
	 * Sets a job-individual threshold to start a distributed calculation.
	 * Please note: Any values below 1000 may very well be counter-productive and make the job slower!
	 * 
	 * @param threshold The number of items that need to be part of the job in order to execute it in distributed mode.
	 * @return
	 */
	public CFSAction<ATTACHEDTO> setDistributedCalculationThreshold(int threshold) {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".setDistributedCalculationThreshold", threshold));
		this.distributedCalculationThreshold = threshold;
		return this;		
	}	
	
	public ATTACHEDTO back() {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".back"));
		return backLink;
	}

	@Override
	public String getObjectName() {
		return backLink.getObjectName() + ".CFSAction";
	}

	/**
	 * Defines subsequent follow-on jobs
	 *
	 * @param jobs List of ChainedJobInfo objects
	 * @return
	 */
	public CFSAction<ATTACHEDTO> setNextJobs(List<ChainedJobInfo> jobs) {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".setNextJobs", jobs));
		this.jobs = jobs;
		return this;
	}

	public String getCfsLabel() { return cfsLabel; }
	public Boolean getCalculate() {	return calculate; }
	public int getDistributedCalculationThreshold() {	return distributedCalculationThreshold; }
	public int getPriority() { return priority; }
	public List<ChainedJobInfo> getNextJobs() { return jobs; }
	
	@Override public String toString() { 
		return "CFSAction [cfsLabel=" + cfsLabel + ", calculate=" + calculate + ", hashCode()=" + hashCode() + "]";
	}

}