package net.pricefx.common.apibuilder.clicoll;

import java.util.List;
import java.util.Map;

import net.pricefx.common.api.HeaderPhase;
import net.pricefx.common.apibuilder.AbstractBuilder;
import net.pricefx.common.apibuilder.IdGenerator;
import net.pricefx.common.apibuilder.Invocations;
import net.pricefx.common.apibuilder.MethodInvocation;

/**
 * This is the shared builder class for Quotes, RebateAgreements and Contracts.<br/>
 * Some important notes on the sequence in which changes to the document structure are processed:<br/><br/>
 * 
 * Any modifying action (adding lines, inputs etc. or a structure object etc.) IS NOT executed as per sequence of the Groovy logic!
 * The Groovy logic is more like a routine to collect all necessary changes. These changes are not executed immediately.
 * They are collected (per execution phase) and then executed in this order:<br/>
 * <ul>
 * <li>Add line items</li>
 * <li>Delete line items (PRE phase only!)</li>
 * <li>Move line items (PRE phase only!)</li>
 * <li>Rename folders</li>
 * <li>Add inputs</li>
 * <li>Add outputs</li>
 * <li>Update header fields</li>
 * </ul>
 * 
 */
@SuppressWarnings("rawtypes")
public abstract class CalculableLineItemCollectionBuilder<T extends CalculableLineItemCollectionBuilder> extends AbstractBuilder {

	final protected HeaderPhase phase;
	
    public static final String METHOD_ADD_OR_UPDATE_OUTPUT = "addOrUpdateOutput";
    public static final String METHOD_ADD_OR_UPDATE_INPUT = "addOrUpdateInput";
    public static final String METHOD_UPDATE_FIELD = "updateField";
    public static final String METHOD_ADD_LINE_ITEM = "addLineItem";    
    public static final String ROOT_LINE_ID = "ROOT";
    
	public CalculableLineItemCollectionBuilder(String instanceId, String classId, IdGenerator idGenerator, Invocations<MethodInvocation> invocations, HeaderPhase phase) {
		super(instanceId, classId, idGenerator, invocations);
		this.phase = phase;
	}

	/**
	 * Determines the current calculation phase of the header logic execution.
	 * 
	 * @return true if the current context is "pre"/before line items calculation
	 */
	public boolean isPrePhase() {
		addInvocation("isPrePhase");
		return HeaderPhase.PREPROCESSING.equals(phase);
	}
	
	/**
	 * Determines the current calculation phase of the header logic execution.
	 * 
	 * @return true if the current context is "post"/after line items calculation
	 */	
	public boolean isPostPhase() {
		addInvocation("isPostPhase");		
		return HeaderPhase.POSTPROCESSING.equals(phase);
	}	

	/**
	 * Adds or updates a result value (an output value) on the ROOT level.
	 *
	 * <p>For map example/details, see {@link #addOrUpdateOutput(String, Map)}
	 *
	 * @param calculationResult A map describing a calculation result
	 * @return itself (for flow pattern)
	 */
    public T addOrUpdateOutput(Map<String, Object> calculationResult) {
        return addOrUpdateOutput(ROOT_LINE_ID, calculationResult);
    }
    
    /**
     * Adds or updates a result value (an output value) on a folder (or line item) with given {@code lineId}
     * 
	 * <p>
	 * Example how to set a background color for the header output results:
	 * <blockquote><pre>
     * quoteprocessor.addOrUpdateOutput(folderId,
     *                  [
     *                      "resultName": "RedMinimumMargin",
     *                      "resultLabel": "Red (Minimum) Margin",
     *                      "result": totalMinimumRed,
     *                      "formatType"   : FieldFormatType.MONEY_EUR,
     *                      "resultType"   : CalculationResultType.SIMPLE,
     *                      "cssProperties": "background-color: red;",
     *                      // "suffix": "",
     *                      // "warnings" : []
     *                  ]
     * )
	 * </pre></blockquote>
	 *
	 * @see net.pricefx.common.api.FieldFormatType
	 * @see net.pricefx.common.api.CalculationResultType
	 *
     * @param lineId The id of the folder/line
	 * @param calculationResult A map describing a calculation result
	 * @return itself (for flow pattern)
	 * 
	 */
	@SuppressWarnings("unchecked")
	public T addOrUpdateOutput(String lineId, Map<String, Object> calculationResult) {
		addInvocation(METHOD_ADD_OR_UPDATE_OUTPUT,lineId, calculationResult);
		return (T) this;
	}
	
	/**
	 * Adds or updates an input parameter on the ROOT level.
	 *
	 * <p>For map example/details, see {@link #addOrUpdateInput(String, Map)}
	 *
	 * @param contextParameter A map describing the context parameter
	 * @return itself (for flow pattern)
	 */
    public T addOrUpdateInput(Map<String, Object> contextParameter) {
        return addOrUpdateInput(ROOT_LINE_ID, contextParameter);
    }
    
    /**
     * Adds or updates an input parameter on a folder (or line) with given {@code lineId}.
	 *
	 * <p>Example:
	 * <blockquote><pre>
	 * if(quoteProcessor.isPrePhase()) {
	 *   quoteProcessor.<strong>addOrUpdateInput</strong>(
	 *       [
	 *         "name"     : "DeliveryType",
	 *         "label"    : "Delivery Type",
	 *         "type"     : InputType.OPTION,
	 *         "required" : true,
	 *         "parameterConfig" : [
	 *           "labels" : [
	 *             "Express" : "Express Delivery",
	 *             "Standard" : "Standard Delivery"
	 *           ]
	 *         ],
	 *         "valueOptions" : [
	 *           "Express",
	 *           "Standard"
	 *         ]
	 *       ]
	 *    )
	 * }
	 * </pre></blockquote>
	 *
	 * <p>Example:
	 * <blockquote><pre>
	 * if (quoteProcessor.isPrePhase()) {
	 *     quoteProcessor.<strong>addOrUpdateInput</strong>(lineItem.lineId, [
	 *        "name": "CompetitorPrice",
	 *        "label": "Competitor Price",
	 *        "type" : InputType.USERENTRY,
	 *        "value": primaryCompetitor,
	 *        "parameterGroup": "Competition Information",
	 *       ])
	 * }
	 * </blockquote></pre>
	 *
	 * <p>Example:
	 * <blockquote><pre>
	 * if(quoteProcessor.isPrePhase()) {
	 *     quoteProcessor.<strong>addOrUpdateInput</strong>( lineId,
	 *           [
	 *              "name"    : "SpecialBundlePromotion",
	 *              "label"   : "SpecialBundlePromotion",
	 *              "type"    : InputType.HIDDEN,
	 *              "required": false,
	 *              "readOnly": true,
	 *              "value"   : value
	 *           ]
	 *     )
	 * }
     * </blockquote></pre>
     *
	 * @see net.pricefx.common.api.InputType
	 *
	 * @param lineId The id of the line
	 * @param contextParameter A map describing the context parameter
	 * @return itself (for flow pattern)
	 * 
	 */
	@SuppressWarnings("unchecked")
	public T addOrUpdateInput(String lineId, Map<String, Object> contextParameter) {
		addInvocation(METHOD_ADD_OR_UPDATE_INPUT,lineId, contextParameter);
		return (T) this;
	}
	
	/**
	 * Updates a field on the line.<br>
	 * Note: Not supported for all objects, e.g. quotes.
	 * 
	 * @param lineId The id of the line
	 * @param fieldName The field name
	 * @param value The new value
	 * @return itself (for flow pattern)
	 */
	@SuppressWarnings("unchecked")
	public T updateField(String lineId, String fieldName, Object value) {
		addInvocation(METHOD_UPDATE_FIELD,lineId, fieldName, value);
		return (T) this;
	}
	
	/**
	 * Updates a header field value. 
	 * <p>
	 * For quotes the supported fields are: <code>additionalInfo1</code>, <code>additionalInfo2</code>, <code>additionalInfo3</code>,
	 * <code>additionalInfo4</code>, <code>externalRef</code>, <code>expiryDate</code>, <code>editabilityStatus</code>, <code>targetDate</code>,
	 * <code>label</code>, <code>userGroupEdit</code>, <code>userGroupViewDetails</code>.
	 * 
	 * @param fieldName The field name
	 * @param value The new value
	 * @return itself (for flow pattern)
	 */
    public T updateField(String fieldName, Object value) {
        return updateField(ROOT_LINE_ID, fieldName, value);
    }
    
    /**
     * Sets (adds or updates) rendering information for particular header fields and buttons.
     * The available key/value pairs depend on a particular field. If a key is used that is unknown to the client, it will be simply ignored.
     * <p>
     * The following keys are currently supported for fields: <code>hide</code>, <code>required</code>, <code>disabled</code> (all boolean type).
     * The field names correspond to the names in the JSON data structure. The following fieldNames are supported:
     * <ul>
	 * <li>Common for Q/CT/RA: <code>uniqueName</code>, <code>workflowState</code>, <code>creationWorkflowStatus</code>, <code>creationWorkflowLabel</code>,
	 * <code>supersededBy</code>, <code>prevRev</code>, <code>label</code>, <code>startDate</code>, <code>endDate</code>, <code>attachments</code>, 
	 * <code>signatureStatus</code>, <code>userGroupEdit</code>, <code>userGroupViewDetails</code>, <code>directLink</code>.</li>
	 * <li>Additionally only for Quotes: <code>externalRef</code>, <code>quoteStatus</code>. </li>
	 * <li>Additionally only for Contracts: <code>targetDate</code>, <code>contractExternalRef</code>, <code>contractStatus</code>.</li>
	 * <li>Additionally only for Rebate Agreements: <code>targetDate</code>, <code>payoutDate</code>, <code>rebateAgreementStatus</code>.</li>
	 * <li>Rebate Agreements Templates have their own list of supported fieldNames: <code>label</code>, <code>startDate</code>, <code>endDate</code>, 
	 * <code>payoutDate</code>, <code>targetDate</code>, <code>userGroupEdit</code>, <code>userGroupViewDetails</code>. </li>
	 * </ul>
	 * As for buttons, the following ones are supported by this method: <code>saveButton</code>, <code>submitButton</code>, <code>deleteButton</code>,
	 * <code>duplicateButton</code>, <code>importExportButton</code>, <code>creationWorkflowSubmitAndNext</code> and <code>creationWorkflowBack</code>.
	 * Buttons only support the key <code>hide</code>.
	 * Once you hide a button, you have to explicitly set hide to false or use {@link #clearRenderInfo(String)} in order to let the button reappear.
	 * 
     * @param fieldName The field or button this setting should be applied on
     * @param key The setting name
     * @param value The setting value
     * @return itself (for flow pattern)
     */
    @SuppressWarnings("unchecked")
	public T setRenderInfo(String fieldName, String key, Object value) {
		addInvocation("setRenderInfo", fieldName, key, value, false);
		return (T) this;
    }

    /**
     * Clears the existing render info for a field or for all fields
     * @param fieldName The field name to clear all info for, or null for all info for all fields
     * @return
     */
    @SuppressWarnings("unchecked")
	public T clearRenderInfo(String fieldName) {
		addInvocation("setRenderInfo", fieldName, null, null, true);
		return (T) this;
    }
    
    /**
     * Adds a new line item in the root folder.
     * @param key The key (usually the sku)
     * @return itself (for flow pattern)
     */
    public T addLineItem(String key) {
        return addLineItem(ROOT_LINE_ID, key);
    }	
    
    /**
     * Adds a new line item in the specified parent (folder ID).
     * @param parentLineId id of parent item
     * @param key The key (usually the sku)
     * @return itself (for flow pattern)
     */
    @SuppressWarnings("unchecked")
	public T addLineItem(String parentLineId, String key) {
		addInvocation(METHOD_ADD_LINE_ITEM,parentLineId, key);
		return (T) this;
	}
    
    /**
     * Adds a new line item in the specified parent (folder ID) and sets some context parameters.
	 * <p>
	 * Sometimes it is required to pass values from a header logic to line items that this logic creates. 
	 * You can use the HIDDEN input type and reference it in the calculation logic.
	 * The following code sample creates a new quote line with parameters fetched from the header logic:
	 * <blockquote>
	 * <pre>
	 * def params = [[name:"MyHiddenParam", type: InputType.HIDDEN, value:"ABC"]]
	 * quoteProcessor.addLineItem("ROOT","MB-0003",params)
	 * </pre>
	 * </blockquote>
	 * In the line calculation logic, this value then can be retrieved by:
	 * <blockquote>
	 * <pre>
	 * def val = api.input("MyHiddenParam")
	 * </pre>
	 * </blockquote>
	 * Notes:
	 * <ul>
	 * <li>This way you can also set default values of "real" input parameters as determined by the line calculation 
	 * logic. Set the value in the header logic and make sure that the names of the parameter match in both logics.</li>
	 * <li>If you use this method only to transport values between the header and line, set the parameter type to HIDDEN. 
	 * The value can also be a map to transport more structured data.</li>
	 * <li>Any parameter that is non-HIDDEN and not determined by the line item logic will be removed.</li>
	 * <li>Be aware that this hidden data is also exposed to the user (it is part of the quote response). 
	 * It is not visible in the UI, but that does not make it a secure storage! Also, try to keep the data volume 
	 * reasonably low.</li>
	 * </ul>
	 * You can also set default values for items created using the quoteStructure class:
	 * <blockquote>
	 * <pre>
	 * def params = [[name:"MyHiddenParam", type: InputType.HIDDEN, value:"ABC"]]
     * def qs = new QuoteStructure()
     * qs.addPart("MySKU",params) //use add.Folder for a folder
	 * </pre>
	 * </blockquote>
     * @param parentLineId id of parent item
     * @param key The key (usually the sku)
     * @param contextParameter A list of maps describing the context parameters
     * @return itself (for flow pattern)
     */
    @SuppressWarnings("unchecked")
	public T addLineItem(String parentLineId, String key, List<Map<String, Object>> contextParameter) {
		addInvocation(METHOD_ADD_LINE_ITEM,parentLineId, key, contextParameter);
		return (T) this;
	}	    
    
    /**
     * Moves an item or folder under a new parent. Only works in the PRE phase.
     * 
     * @param lineId The line ID of the item to move
     * @param newParentId The line ID of the target folder. null if you want to move it to root folder
     * @return itself (for flow pattern)
     */
    @SuppressWarnings("unchecked")
	public T moveItem(String lineId, String newParentId) {
		addInvocation("moveItem",lineId, "parentId", newParentId);
		return (T) this;		
	}
	
    /**
     * Renames a folder (not items).
     * @param lineId The line ID of the folder to rename
     * @param newFolderLabel New label
     * @return itself (for flow pattern)
     */
    @SuppressWarnings("unchecked")
	public T renameFolder(String lineId, String newFolderLabel) {
		addInvocation("renameFolder",lineId, "label", newFolderLabel);
		return (T) this;	
	}
	
    /**
     * Deletes an item or folder. Only works in the PRE phase.
     * @param lineId The line ID of the item to remove
     * @return itself (for flow pattern)
     */
    @SuppressWarnings("unchecked")
	public T deleteItem(String lineId) {
		addInvocation("deleteItem",lineId);
		return (T) this;	
	}    
}
