package net.pricefx.formulaengine;

import net.pricefx.formulaengine.scripting.Matrix2D;
import net.pricefx.formulaengine.scripting.PublicGroovyAPI;
import net.pricefx.server.dto.calculation.ResultMatrix;

import java.sql.SQLException;
import java.util.Collection;
import java.util.List;
import java.util.Map;

/**
 * Represents the data of a PA field collection (Data Feed, Data Source, Datamart, etc.) and provides methods to work with it.
 *
 * Available in a PA DataLoad context only. If you need an alternative, see {@link DatamartContext.DataLoader}.
 * In order to get a valid instance of {@link DatamartRowSet}, use <code>api.getDatamartRowSet()</code>. Typically only
 * changes to the <b>target</b> {@link DatamartRowSet} will actually be stored. And which field collection is actually
 * the source or the target depends on the Dataload configuration.
 *
 * <p><b>Example:</b></p>
 * <blockquote>
 *     <pre>
 *         def sourceRows = api.getDatamartRowSet("source");
 *         def targetRows = api.getDatamartRowSet("target");
 *     </pre>
 * </blockquote>
 *
 * <ul>
 * <li>see {@link PublicGroovyAPI#getDatamartRowSet()}</li>
 * <li>see {@link PublicGroovyAPI#getDatamartRowSet(String)}</li>
 * </ul>
 *
 */
public interface DatamartRowSet {

    int RESULTMATRIX_MAXROWS = 100;

    /**
     * Provides the list of the key fields.
     *
     * @return key field names represented as a {@link List} of {@link String}
     */
    List<String> getKeyFields();

    /**
     * Provides the list of all the fields.
     *
     * @return field names represented as a {@link Collection} of {@link String}
     */
    Collection<String> getAllFields();


    /**
     * Adds a row.
     *
     * <blockquote><pre>
     *      target.addRow([
     *                 sku: obj.sku,              // key field
     *                 salesOrg: obj.attribute1,  // key field
     *                 quantity: 0
     *         ])
     * </pre></blockquote>
     *
     * @param map represented as {@link Map}
     * @return true if the row was added
     */
    boolean addRow(Map<String,Object> map);

    /**
     * Updates row values, only if a row already exists for the provided keys.
     *
     * <blockquote><pre>
     *      target.updateRow([
     *                 sku: obj.sku,              // key field
     *                 salesOrg: obj.attribute1,  // key field
     *                 quantity: 0
     *         ])
     * </pre></blockquote>
     *
     * @param map row represented as {@link Map}
     * @return true if the row was updated
     */
    boolean updateRow(Map<String,Object> map);

    /**
     * Deletes a row.
     *
     * Note that only the key field values are relevant, the method does not verify that every single value matches.
     *
     * <blockquote><pre>
     *      target.deleteRow([
     *                 txid: obj.txid
     *         ])
     * </pre></blockquote>
     *
     * @param map key fields represented as {@link Map}
     * @return true if the row was deleted
     */
    boolean deleteRow(Map<String,Object> map);

    /**
     * Get the row with the given key values.  Always returns a copy.
     *
     * <blockquote><pre>
     *     def rowKey = [
     *         "SHIP_TO": obj.ship_to,
     *         "SALES_MANAGER": obj.sales_manager,
     *         "MATERIAL": obj.sku,
     *         "REPORT_MONTH": prPeriod]
     *     def rRow = target.getRow(rowKey)
     *
     * @param map key fields represented as {@link Map}
     * @return row represented as {@link Map}
     */
    Map<String, Object> getRow(Map<String,Object> map);

    /**
     * Set default values to be used when creating new rows.
     *
     * This method will drop all previously set defaults. To add to the existing defaults, use {@link #addDefaultValues(Map)}
     *
     * @see #newRow()
     * @see #addDefaultValues(Map)
     */
    void setDefaultValues(Map<String,Object> defaultValues);

    /**
     * Add default values to be used when creating new rows.
     *
     * There can be only one default value per field, so previously set defaults may be replaced.
     *
     * @see #newRow()
     * @see #setDefaultValues(Map)
     */
    void addDefaultValues(Map<String,Object> defaultValues);

    /**
     * Creates a representation of a new row, potentially populated with default values.
     *
     * Note that the row is not yet added to the {@link DatamartRowSet}. Use {@link #addRow(Map)} to add the row.
     *
     * @see #addRow(Map)
     * @see #setDefaultValues(Map)
     * @see #addDefaultValues(Map)
     * @return new row represented as {@link Map}
     */
    Map<String, Object> newRow();

    /**
     * Adds rows from another {@link DatamartRowSet}.
     *
     * Note that there is no check whether the fields of the rowsets are compatible, not even regarding the key fields.
     *
     * @param otherRowset {@link DatamartRowSet} containing the rows
     * @throws IllegalFieldValuesException if the key fields are not the same
     */
    void loadRows(final DatamartRowSet otherRowset) throws Exception;


    /**
     * Provides a <code>Collection</code> of <code>Map</code> representing every row in this {@link DatamartRowSet}.
     *
     * @return a <code>Collection</code> of all rows, each row represented as a #Map
     */
    Collection<? extends Map<String, Object>> getRows();

    /**
     * Returns <code>true</code> if this {@link DatamartRowSet} contains no rows.
     *
     * @return <code>true</code> if this {@link DatamartRowSet} contains no rows
     */
    boolean isEmpty();

    /**
     * Returns the number of rows in this {@link DatamartRowSet}.
     *
     * If this {@link DatamartRowSet} contains more than <code>Integer.MAX_VALUE</code> elements, returns <code>Integer.MAX_VALUE</code>.
     *
     * @return the number of rows in this {@link DatamartRowSet}
     */
    int size();

    /**
     * Resets the current row index of this {@link DatamartRowSet}.
     *
     * @see #first()
     * @see #next()
     * @see #getCurrentRow()
     */
    void reset();

    /**
     * Sets the current row index of this {@link DatamartRowSet} to the first row.
     *
     * @see #reset()
     * @see #next()
     * @see #getCurrentRow()
     */
    boolean first() throws SQLException;

    /**
     * Sets the current row index of this {@link DatamartRowSet} to the last row.
     *
     * @see #reset()
     * @see #first()
     * @see #next()
     * @see #getCurrentRow()
     */
    boolean last() throws SQLException;

    /**
     * Sets the current row index of this {@link DatamartRowSet} to the previous row.
     *
     * @see #reset()
     * @see #first()
     * @see #next()
     * @see #getCurrentRow()
     */
    boolean previous() throws SQLException;

    /**
     * Sets the current row index of this {@link DatamartRowSet} to the next row.
     *
     * @see #reset()
     * @see #first()
     * @see #getCurrentRow()
     */
    boolean next () throws SQLException;

    /**
     * Returns the current row.
     *
     * @see #reset()
     * @see #first()
     * @see #next()
     * @return current row represented as {@link Map}
     */
    Map<String, Object> getCurrentRow();

    /**
     * Deletes the current row.
     * @return true if successful (row existed)
     */
    boolean deleteCurrentRow();

    /**
     * Updates rows in this set, for which a row with the same key exists in the other set.
     * Does not add any rows to this set!
     *
     * @param otherSet other {@link DatamartRowSet}
     */
    void merge(final DatamartRowSet otherSet);

    /**
     * Add rows from a #Matrix2D.
     *
     * If the #Matrix2D data does not contain all {@link DatamartRowSet} key fields, then we assume auto-increment of the missing key fields is desired.
     * If on the other hand, the fields are there, but the values are missing (for some rows), then we assume inserting empty #String <CODE>("")</CODE>.
     *
     * @see #loadRows(Matrix2D, Collection)
     * @see #loadRows(Matrix2D, Map)
     * @param data the #Matrix2D holding the data to add
     */
    void loadRows(final Matrix2D data);

    /**
     * Add rows from a #Matrix2D, possibly only a subset of the matrix columns.
     *
     * If the #Matrix2D data does not contain all {@link DatamartRowSet} key fields, then we assume auto-increment of the missing key fields is desired.
     * If on the other hand, the fields are there, but the values are missing (for some rows), then we assume inserting empty #String <CODE>("")</CODE>.
     *
     * @see #loadRows(Matrix2D)
     * @see #loadRows(Matrix2D, Map)
     * @param data the #Matrix2D holding the data to add
     * @param columns name of the #Matrix2D columns to import
     */
    void loadRows(final Matrix2D data, final Collection<String> columns);

    /**
     * Add rows from a #Matrix2D, possibly only a subset of the matrix columns.
     *
     * If the #Matrix2D data does not contain all {@link DatamartRowSet} key fields, then we assume auto-increment of the missing key fields is desired.
     * If on the other hand, the fields are there, but the values are missing (for some rows), then we assume inserting empty #String <CODE>("")</CODE>.
     *
     * @see #loadRows(Matrix2D)
     * @see #loadRows(Matrix2D, Collection)
     * @param data the #Matrix2D holding the data to add
     * @param fieldMapping key is the {@link DatamartRowSet} field name, and value is the Matrix column name
     */
    void loadRows(final Matrix2D data, final Map<String,String> fieldMapping);

    /**
     * Converts the data to {@link ResultMatrix}
     *
     * Limited to {@link #RESULTMATRIX_MAXROWS} rows.
     *
     * @return converted to {@link ResultMatrix}
     */
    ResultMatrix toResultMatrix();

    /**
     * Converts the data to {@link ResultMatrix}
     *
     * Limited to user defined number of rows which cannot exceed {@link #RESULTMATRIX_MAXROWS}.
     *
     * @param maxRows maximum number of rows to be converted, capped to {@link #RESULTMATRIX_MAXROWS}
     * @return converted to {@link ResultMatrix}
     */
    ResultMatrix toResultMatrix(final int maxRows);

}
